include "programming_language.dfy"
include "assertions.dfy"
include "inference_rules.dfy"
include "concurrent_example_spec.dfy"
include "util.dfy"

module concurrent_example_proof {
    import opened programming_language
    import opened assertions
    import opened inference_rules
    import opened concurrent_example_spec
    import opened util

    // ======
    // Step 1
    // ======
    lemma assignment_axiom()
        ensures SequentiallyDerivable(p4(), C(), q4())
    {
        assert OwnFromSeq(["x"]) == Own("x");
    }


    // ======
    // Step 2
    // ======
    lemma frame_rule()
        ensures SequentiallyDerivable(p3(), C(), q3())
    {
        assignment_axiom();
    }

    // ======
    // Step 3
    // ======
    lemma helper_CR_1()
        ensures AssertionImpl(p2(), p3())
        ensures AssertionImpl(q3(), q2())

    lemma consequence_rule_1()
        ensures SequentiallyDerivable(p2(), C(), q2())
    {
        frame_rule();
        helper_CR_1();
    }

    // ======
    // Step 4
    // ======
    lemma auxiliary_variable_elimination()
        ensures SequentiallyDerivable(p1(), C(), q1())
    {
        consequence_rule_1();
    }

    // ======
    // Step 5
    // ======
    lemma helper_CR_2()
        ensures AssertionImpl(p(), p1())
        ensures AssertionImpl(q1(), q())

    lemma consequence_rule_2()
        ensures SequentiallyDerivable(p(), C(), q())
    {
        auxiliary_variable_elimination();
        helper_CR_2();
    }


    // ======
    // Step 6
    // ======

    lemma helper_condition()
        ensures forall m :: SemanticAssertion(p(), m) ==> DefinedBExpr(True, m.0, m.1)
    {
    }

    lemma helper_Sta()
        ensures Sta(p(), Action.Indep(R(), Id()))
        ensures Sta(q(), Action.Indep(R(), Id()))

    lemma helper_AndTrue()
        requires SequentiallyDerivable(p(), C(), q())
        ensures SequentiallyDerivable(Assertion.And(p(), Bool(True)), C(), q());

    lemma helper_guarantee()
        ensures ActionImpl(Change(p(), q()), Action.Indep(G(), TrueAct()))

    lemma helper_invariant()
        ensures AssertionImpl(Assertion.Or(p(), q()), Assertion.Indep(I(), Bool(True)))

    lemma helper_fence()
        ensures Fence(I(), R())
        ensures Fence(I(), G())

    lemma ATOM_rule()
        ensures ConcurrentlyDerivable(R(), G(), I(), p(), C'(), q())
    {
        helper_condition();
        helper_guarantee();
        helper_invariant();
        helper_fence();
        helper_Sta();
        consequence_rule_2();
        helper_AndTrue();
    }


// */
}
