include "programming_language.dfy"
include "assertions.dfy"
include "inference_rules.dfy"
include "util.dfy"

module gcd_program_proof {
    import opened programming_language
    import opened assertions
    import opened inference_rules
    import opened util


    function p0(): Assertion {
        Assertion.And(
            OwnFromSeq(["x", "y"]),
            Assertion.And(
                Bool(Equal(LVar("X"), Int(7))),
                Bool(Equal(LVar("Y"), Int(5)))
            )
        )
    }

    function p1(): Assertion {
        Assertion.And(
            OwnFromSeq(["x", "y"]),
            Assertion.And(
                Bool(Equal(Var("x"), LVar("X"))),
                Bool(Equal(LVar("Y"), Int(5)))
            )
        )
    }

    function p2(): Assertion {
        Assertion.And(
            OwnFromSeq(["x", "y"]),
            Assertion.And(
                Bool(Equal(Var("x"), LVar("X"))),
                Bool(Equal(Var("y"), LVar("Y")))
            )
        )
    }

    function p_x_assigment(): Assertion {
        Assertion.And(
            OwnFromSeq(["x"]),
            Assertion.And(
                Bool(Equal(LVar("X"), Int(7))),
                emp_h
            )
        )
    }

    function q_x_assigment(): Assertion {
        Assertion.And(
            OwnFromSeq(["x"]),
            Assertion.And(
                Bool(Equal(Var("x"), LVar("X"))),
                emp_h
            )
        )
    }

    function p_y_assigment(): Assertion {
        Assertion.And(
            OwnFromSeq(["y"]),
            Assertion.And(
                Bool(Equal(LVar("Y"), Int(5))),
                emp_h
            )
        )
    }

    function q_y_assigment(): Assertion {
        Assertion.And(
            OwnFromSeq(["y"]),
            Assertion.And(
                Bool(Equal(Var("y"), LVar("Y"))),
                emp_h
            )
        )
    }

    function x_mem(): Assertion {
        Assertion.And(Own("x"), Bool(Equal(Var("x"), LVar("X"))))
    }

    function y_mem(): Assertion {
        Assertion.And(Own("y"), Bool(Equal(LVar("y"), Int(5))))
    }


    function x_command(): Stmt {
        Assign("x", Int(7))
    }
    function y_command(): Stmt {
        Assign("y", Int(5))
    }


    lemma x_assignment_axiom()
        ensures SequentiallyDerivable(p_x_assigment(), x_command(), q_x_assigment())
    {
    }

    lemma y_assignment_axiom()
        ensures SequentiallyDerivable(p_y_assigment(), y_command(), q_y_assigment())
    {
    }

    lemma x_frame_rule()
        ensures SequentiallyDerivable(
            Assertion.Indep(p_x_assigment(), y_mem()),
            x_command(),
            Assertion.Indep(q_x_assigment(), y_mem())
        )
    {
        x_assignment_axiom();
    }

    lemma y_frame_rule()
        ensures SequentiallyDerivable(
            Assertion.Indep(p_y_assigment(), x_mem()),
            y_command(),
            Assertion.Indep(q_y_assigment(), x_mem())
        )
    {
        y_assignment_axiom();
    }

    lemma helper_x_implication()
        ensures AssertionImpl(
            p0(),
            Assertion.Indep(p_x_assigment(), y_mem())
        )
        ensures AssertionImpl(
            Assertion.Indep(q_x_assigment(), y_mem()),
            p1()
        )

    lemma x_consequence_rule()
        ensures SequentiallyDerivable(p0(), x_command(), p1())
    {
        x_frame_rule();
        helper_x_implication();
    }

    lemma helper_y_implication()
        ensures AssertionImpl(
            p1(),
            Assertion.Indep(p_y_assigment(), x_mem())
        )
        ensures AssertionImpl(
            Assertion.Indep(q_y_assigment(), x_mem()),
            p2()
        )

    lemma y_consequence_rule()
        ensures SequentiallyDerivable(p1(), y_command(), p2())
    {
        y_frame_rule();
        helper_y_implication();
    }

    lemma rule_of_composition()
        ensures SequentiallyDerivable(
            p0(),
            Seq(x_command(), y_command()),
            p2()
        )
    {
        x_consequence_rule();
        y_consequence_rule();
    }
}
