include "programming_language.dfy"
include "assertions.dfy"
include "util.dfy"
include "decreasing_functions.dfy"

module inference_rules {
    import opened programming_language
    import opened assertions
    import opened util
    import decreasing_functions

    predicate wffAct(R: RelAct, D: DAct) {
        forall mem0, mem1, t ::
            mem_check_t(mem0, t) && mem_check_t(mem1, t) && Semantic_RelAct(R, mem0, mem1)
            ==>
            Semantic_DAct(Bracket(D), mem0, mem1)
            && (forall t' :: t != t' ==>
                Semantic_DAct(Bracket(D), mem_set_t(mem0, t'), mem_set_t(mem1, t'))
                || Semantic_DAct(D, mem_set_t(mem0, t'), mem_set_t(mem1, t')))
    }

    predicate mem_check_t(mem: AllMem, t: int) {
        mem_check_var(mem, "t", t)
    }

    function mem_set_t(mem: AllMem, t:int):AllMem {
        match mem
        case (s0, h0, s1, h1) => (s0["t" := t], h0, s1, h1)
    }

    predicate Sta(p: FullAssn, R: RelAct)
    {
        forall mem0, mem1, u, w, C ::
            u >= 0 && w >= 0
            && Semantic_FullAssn(p, mem0, u, w, C)
            && Semantic_RelAct(R, mem0, mem1)
            ==>
            Semantic_FullAssn(p, mem1, u, w, C)
    }

    predicate Arrow(p: FullAssn, q: FullAssn) {
        forall t: int, con, abs, u, w, C, abs_F :: u >= 0 &&  w >= 0
            && Semantic_FullAssn(p, MemToAllMem(con, abs), u, w, C)
            && MemUnionable(abs, abs_F)
            ==>
            (exists C', abs' ::
                true /* TODO */
                && Semantic_FullAssn(q, MemToAllMem(con, abs'), u, w, C')
            )
    }


    predicate DefiniteProgress(J: RelAssn, R: RelAct, G: RelAct,
            D: DAct, f: (int, AllMem) -> nat, Q: RelAssn)
        requires forall t, mem :: f.reads(t, mem) == {} && f.requires(t, mem)
    {
        forall t, mem :: mem_check_t(mem, t) && Semantic_RelAssn(J, mem)
        ==>
        Semantic_RelAssn(Q, mem)
        || (exists t' :: t' != t && Semantic_RelAssn(Enabled(D), mem_set_t(mem, t')))
        || (forall t', mem' :: t' != t && mem_check_t(mem', t)
            && Semantic_RelAct(R, mem, mem)
            && Semantic_RelAct(AngleBracket(D),mem_set_t(mem, t'), mem_set_t(mem', t'))
            ==>
            f(t, mem') < f(t, mem))
        || (forall mem' :: mem_check_t(mem', t)
            && Semantic_RelAct(RelAct.Or(R, G), mem, mem')
            ==>
            f(t, mem') <= f(t, mem))
    }


    predicate object_is_derivable(D: DAct, R: RelAct, G: RelAct,
            P: RelAssn, con: ODecl, abs: ODecl)
    {
        domain(con) == domain(abs)
        && (forall f :: f in domain(con) ==> (
            exists x, y, C, C' :: con[f] == (x, C) && abs[f] == (y, C')
            && lili_is_derivable(D, R, G,
                FullAssn.And(FullAssn.'(P), FullAssn.And(FullAssn.'(Bool(Equal(Var(x), Var(y)))), Arem(C))),
                C,
                FullAssn.And(FullAssn.'(P), Arem(Skip)))))
        && (forall t, t', m0, m1 :: mem_check_t(m0, t) && mem_check_t(m1, t) && t != t' ==>
            (Semantic_RelAct(G, m0, m1) ==> Semantic_RelAct(R, mem_set_t(m0, t'), mem_set_t(m1, t'))))
        && wffAct(R, D)
        && (forall m :: Semantic_RelAssn(P, m) ==> ! Semantic_RelAssn(Enabled(D), m))
    }


    inductive predicate lili_is_derivable(D: DAct, R: RelAct, G: RelAct,
            p: FullAssn, C: Stmt, q:FullAssn) {

        match C
        // WHL
        case While(B, C') => (
            q == FullAssn.And(p, FullAssn.'(RelAssn.Bool(Not(B))))
            && exists f, p', J, D', Q :: (
                f in decreasing_functions.get_list()
                && helper_impl_FullAssn(FullAssn.And(p, FullAssn.'(Bool(B))), q)
                && helper_impl_RelAssn(ToRelAssn(p), J)
                && helper_impl_FullAssn(
                    FullAssn.And(
                        FullAssn.And(p, FullAssn.'(Bool(B))),
                        FullAssn.'(RelAssn.Or(Enabled(D), Q))
                    ),
                    FullAssn.Indep(p', FullAssn.And(White(Nat(1)), FullAssn.'(Emp)))
                )
                && lili_is_derivable(D, R, G, p', C', p)
                && Sta(FullAssn.'(J), RelAct.Or(R, G))
                && LeqDAct(D', D)
                && wffAct(R, D')
                && DefiniteProgress(J, R, G, D', f, Q)
            )
        )
        case Atomic(C') => (
            // ATOM
            (
                R == Id()
                && (exists q' :: sep_is_derivable(p, C', q')
                    && Arrow(q', q)
                    && helper_impl_RelAct(RelAct.Action(ToRelAssn(p), ToRelAssn(q)), G))
            )
            // ATOM-R
            || (lili_is_derivable(D, Id(), G, p, C', q) && Sta(p, R) && Sta(q, R))
        )
        // RET
        case Return(E) => (
            exists p', E' ::
                && p == FullAssn.And(IgnoreStmt(p'), Arem(Return(E')))
                && q == FullAssn.And(IgnoreStmt(p'), Arem(Skip))
                && helper_impl_FullAssn(p, FullAssn.'(Bool(Equal(E, E'))))
                && Sta(p', R)
        )
        // SEQ
        case Compose(C1, C2) => (
            exists r :: lili_is_derivable(D, R, G, p, C1, r)
                && lili_is_derivable(D, R, G, r, C2, q)
        )
        // IF
        case Cond(B, C1, C2) => (
            lili_is_derivable(D, R, G, FullAssn.And(p, FullAssn.'(Bool(B))), C1, q)
            && lili_is_derivable(D, R, G, FullAssn.And(p, FullAssn.'(Bool(Not(B)))), C2, q)
        )
        // else (ignored) =============
        case Instr(_) => false
        case Call(_, _, _) => false
        case End => false
        case Skip => false
        // end of ignored =============

        // HIDE-◊
        || (
            exists p', q' :: p == IgnoreWhite(p') && q == IgnoreWhite(q')
                && lili_is_derivable(D, R, G, p', C, q')
        )
        // AREM
        || (
            exists p', C1, C2, C3 ::
                p == FullAssn.And(IgnoreWhite(p'), Arem(Compose(C1, C3)))
                && q == FullAssn.And(IgnoreWhite(p'), Arem(Compose(C2, C3)))
                && lili_is_derivable(D, R, G,
                    FullAssn.And(IgnoreWhite(p'), Arem(C1)),
                    C,
                    FullAssn.And(IgnoreWhite(p'), Arem(C2))
                )
         )
        // CSQ
        || (
            exists p', q', R', G' ::
                helper_impl_FullAssn(p, p')
                && helper_impl_FullAssn(q', q)
                && helper_impl_RelAct(R, R')
                && helper_impl_RelAct(G', G)
                && lili_is_derivable(D, R', G', p', C, q')
                && Sta(p, R)
                && Sta(q, R)
                && wffAct(R', D)
        )
        // EX
        || (
            exists p', q', x ::
                p == FullAssn.ForAll(x, p')
                && q == FullAssn.ForAll(x, q')
                && lili_is_derivable(D, R, G, p', C, q')
                && x !in fv_DAct(D)
                && x !in fv_RelAct(R)
                && x !in fv_RelAct(G)
        )
        // CONJ
        || (
            exists p1, p2, q1, q2 ::
                p == FullAssn.And(p1, p2)
                && q == FullAssn.And(q1, q2)
                && lili_is_derivable(D, R, G, p1, C, q1)
                && lili_is_derivable(D, R, G, p2, C, q2)
        )
        // DISJ
        || (
            exists p1, p2, q1, q2 ::
                p == FullAssn.Or(p1, p2)
                && q == FullAssn.Or(q1, q2)
                && lili_is_derivable(D, R, G, p1, C, q1)
                && lili_is_derivable(D, R, G, p2, C, q2)
        )
    }

    predicate helper_impl_FullAssn(p: FullAssn, q: FullAssn) {
        forall m, u, w, C ::
            u >= 0 && w >= 0 && Semantic_FullAssn(p, m, u, w, C)
            ==>
            Semantic_FullAssn(q, m, u, w, C)
    }
    predicate helper_impl_RelAssn(P: RelAssn, Q: RelAssn) {
        forall m :: Semantic_RelAssn(P, m) ==> Semantic_RelAssn(Q, m)
    }
    predicate helper_impl_RelAct(R: RelAct, G: RelAct) {
        forall m0, m1 :: Semantic_RelAct(R, m0, m1) ==> Semantic_RelAct(G, m0, m1)
    }


    predicate sep_is_derivable(p: FullAssn, C: Stmt, q: FullAssn) {
        true
    }
}
